#!/usr/bin/env python

'''Service Directory for SSS'''
__revision__ = '$Id$'

from elementtree.ElementTree import Element, SubElement, XML, tostring
from getopt import getopt, GetoptError
from os import close, fork
from select import select
from signal import signal, SIGINT
from sys import argv, exc_info
from syslog import syslog, LOG_INFO, LOG_ERR
from time import sleep, strftime, localtime
from traceback import extract_tb

from sss.restriction import Data, DataSet
from sss.server import Server, BufferedPipe
from sss.ssslib import comm_lib, InitError, ConnectError, CommunicationError

class Location(Data):
    '''Datatype for Service Locations'''
    def __init__(self, element):
        self.element = element
        for field in ['tier', 'schema_version']:
            if not self.element.attrib.get(field, None):
                self.element.attrib[field] = 1

class service_directory(Server):
    '''Service Directory component'''
    __implementation__ = 'sd'
    __component__ = 'service-directory'
    __statefields__ = ['locations']
    __dispatch__ = {'add-location':'AddLocation',
                    'get-location':'locations.Get',
                    'del-location':'locations.Del'}
    __validate__ = 0

    def __setup__(self):
        self.locations = DataSet("locations", "location", Location, None, False)
        (port, host, proto, sversion, tier)=[comm_lib().GetConfig().get(field) for field in 
                                            ['sd_port', 'sd_host', 'sd_protocol', 'sd_schema_version', 'sd_tier']]
        if [field for field in [port, host, proto, sversion, tier] if field == None]:
            print "Insufficient data in sss.conf.  Exiting..."
            raise SystemExit, 1
        myloc = XML('''<add-location><location component='service-directory' host='%s' port='%s' protocol='%s' schema_version='%s' tier='%s'/></add-location>''' % (host, port, proto, sversion, tier))
        self.locations.Add(myloc, ('localhost', 0))
        self.port = int(port)
        self.protocol = proto

    def AddLocation(self, xml, (peer, port)):
        '''Add new location'''
        return self.locations.Add(xml, (peer, port), self.SendLocEvt_cb)

    def SendLocEvt_cb(self, loc, arg):
        '''Callback for AddLocation to send registration events'''
        if loc.element.attrib['component'] != 'event-manager':
            self.regevt.WriteMessage(("service-directory", "service-registered",
                                      loc.element.attrib['component']))
        
    def ChildSigHand(self, signum, frame):
        '''Signal Handler for Children'''
        raise SystemExit

    def __setupChildren__(self):
        '''Setup child processes'''
        self.regevt = BufferedPipe()
        self.pipes.append(self.regevt)
        pid = fork()
        if pid == 0:
            signal(SIGINT, self.ChildSigHand)
            msgs = []
            comm = comm_lib()
            while 1:
                (infd, out, x) = select([self.regevt.r], [], [], 5)
                if infd != []:
                    self.regevt.ReadMessages()

                while self.regevt.buffer:
                    try:
                        msgs.append(self.regevt.ReadMessage())
                    except MsgError:
                        break

                if msgs:
                    e = Element("add-event")
                    t = strftime("%c", localtime())
                    for (c, m, d) in msgs:
                        SubElement(e, "event", component=c, msg=m, data=d, time=t)
                    try:
                        h = comm.ClientInit("event-manager")
                        comm.SendMessage(h, tostring(e))
                        comm.RecvMessage(h)
                        comm.ClientClose(h)
                        msgs = []
                    except:
                        syslog(LOG_ERR, "Failed to send registration event")
                    del e
        else:
            pinger_pid = fork()
            if pinger_pid == 0:
                self.RunPinger()
            else:
                close(self.regevt.r)
                return [pid, pinger_pid]

    def RunPinger(self):
        try:
            c = comm_lib()
            while 1:
                sleep(300) # 5 minutes
                h = c.ClientInit("service-directory")
                c.SendMessage(h, "<get-location><location component='*' host='*' port='*' protocol='*'/></get-location>")
                locs = XML(c.RecvMessage(h)).findall('location')
                c.ClientClose(h)
                for loc in locs:
                    comp = loc.attrib["component"]
                    try:
                        h = c.ClientInit(comp)
                        c.ClientClose(h)
                    except (ConnectError, CommunicationError):
                        h = c.ClientInit("service-directory")
                        c.SendMessage(h, "<del-location><location component='%s'/></del-location>" % comp)
                        syslog(LOG_INFO, "component %s flushed from sd due to connect error" % comp)
                        c.RecvMessage(h)
                        c.ClientClose(h)
                    except:
                        (t, v, tb) = exc_info()
                        syslog(LOG_ERR, "Failure in sd-pinger")
                        syslog(LOG_ERR, "Attempted to contact %s" % comp)
                        for line in extract_tb(tb):
                            errstr = '  File "%s", line %i, in %s\n    %s\n' % line
                            syslog(LOG_ERR, errstr)
                            errstr = "%s: %s\n" % (t, v)
                            syslog(LOG_ERR, errstr)
                        del t, v, tb
        except KeyboardInterrupt:
            raise SystemExit, 1
        except:
            (t, v, tb) = exc_info()
            syslog(LOG_ERR, "*** unexpected Failure in sd-pinger")
            for line in extract_tb(tb):
                errstr = '  File "%s", line %i, in %s\n    %s\n' % line
                syslog(LOG_ERR, errstr)
            errstr = "%s: %s\n" % (t, v)
            syslog(LOG_ERR, errstr)
            del t, v, tb
            raise SystemExit, 1

if __name__ == "__main__":
    try:
        (opts, args) = getopt(argv[1:], 'd', ['daemon='])
    except GetoptError, msg:
        print "%s\nUsage:\nsd.py [-d] [--daemon <pidfile>]" % (msg)
        raise SystemExit, 1
    daemon = [opt for opt in opts if opt[0] == '--daemon']
    ldebug = len([opt for opt in opts if opt[0] == '-d'])
    if daemon:
        from sss.daemonize import daemonize
        daemonize(daemon[0][1])
    try:
        s = service_directory(debug=ldebug)
    except InitError, msg:
        print "Library init Error:%s" % (msg)
        raise SystemExit, 1
    s.ServeForever()
