'''This module implements the Less Restricted Syntax for the SSS Project'''
__revision__ = '$Revision$'

from re import compile as regcompile
from syslog import syslog, LOG_INFO
from elementtree.ElementTree import Element

class LRFieldError(Exception):
    '''Error raised in case of field exception'''
    pass

class LRData(object):
    '''LR Data are Less-Restricted Syntax Objects'''
    def __init__(self, element):
        self.element = element
        self.fields = [node.tag for node in self.element.getchildren()]

    def get(self, field):
        '''fetch a single-instance attribute'''
        if field not in self.fields:
            raise LRFieldError
        return self.element.find(field).text

    def set(self, field, value):
        '''set a single-instance attribute'''
        if field not in self.fields:
            raise LRFieldError
        self.element.find(field).text = value

    def MatchField(self, field, spec):
        '''Single Field Match'''
        negate = spec.get('negate', 'false')
        operation = spec.get('op', 'eq')
        test = False
        if ((operation == 'eq') and (field.text == spec.text)):
            test = True
        elif ((operation == 'lt') and (int(field.text) < int(spec.text))):
            test = True
        elif ((operation == 'gt') and (int(field.text) > int(spec.text))):
            test = True
        elif ((operation == 'range') and
              (int(spec.text.split('-')[0]) <= int(field.text) <= int(spec.text.split('-')[1]))):
            test = True
        elif ((operation == 're') and regcompile(spec.text).match(field.text)):
            test = True
        if negate == 'false':
            return test
        else:
            return not test

    def Match(self, other):
        '''Perform Full Data Match'''
        if self.element.tag != other.tag:
            return False
        for child in [field for field in other.getchildren() if field.get('match', 'true') == 'true']:
            if self.element.find(child.tag):
                if not self.MatchField(self.element.find(child.tag), child):
                    return False
        return True

    def Display(self, other):
        '''Return an object matching the specification'''
        new = Element(self.element.tag)
        for child in [field for field in other.getchildren() if field.get('return', 'true') == 'true']:
            if self.element.find(child.tag) != None:
                new.append(self.element.find(child.tag))
        return new

class LRDataSet(object):
    '''This class provides a list store for Data objects, complete with functions that do most of the
    restriction syntax work
    rname is the container name used in responses
    ename is the element name used for specifications
    obj is the constructor to call on new instances for the datastore
    idalloc is a unique id allocator; must be paired with self.obj function signature
    log is if add/del operations should be logged.'''
    def __init__(self, rname, ename, obj, idalloc, log):
        self.data = []
        self.rname = rname
        self.ename = ename
        self.obj = obj
        self.idalloc = idalloc
        self.log = log

    def __iter__(self):
        return iter(self.data)

    def remove(self, item):
        '''remove a instance object'''
        return self.data.remove(item)

    def append(self, item):
        '''add a new instance object'''
        return self.data.append(item)
    
    def cb_Add(self, spec, portinfo, callback=None):
        '''Implement restriction syntax Add for component use'''
        ret = Element(self.rname)
        for espec in spec.findall(self.ename):
            if self.idalloc:
                child = self.obj(espec, self.idalloc.get())
            else:
                child = self.obj(espec)
            if self.log:
                syslog(LOG_INFO, "Got %s from %s"%(self.ename, portinfo[0]))
            self.data.append(child)
            if callback:
                callback(child, spec.attrib)
            ret.append(child.Display(espec))
        return ret

    def cb_Get(self, spec, portinfo, callback=None):
        '''Implement restriction syntax Get for component use'''
        ret = Element(self.rname)
        for espec in spec.findall(self.ename):
            for item in [datum for datum in self.data if datum.Match(espec)]:
                if callback:
                    callback(item, spec.attrib)
                ret.append(item.Display(espec))
        return ret

    def Del(self, spec, portinfo, callback=None):
        '''Implement restriction syntax Del for component use'''
        ret = Element(self.rname)
        for espec in spec.findall(self.ename):
            for item in [datum for datum in self.data if datum.Match(espec)]:
                self.data.remove(item)
                if self.log:
                    syslog(LOG_INFO, "Got %s from %s" % (spec.tag, portinfo[0]))
                if callback:
                    callback(item, spec.attrib)
                ret.append(item.Display(espec))
        return ret

class LRSizedSet(LRDataSet):
    '''SizedSet is a DataSet instance that rotates through data, based on min and max'''
    def __init__(self, rname, ename, obj, idalloc, log, smin, smax):
        LRDataSet.__init__(self, rname, ename, obj, idalloc, log)
        self.min = smin
        self.max = smax
        
    def Add(self, spec, (peer, port), callback=None):
        '''Add a new entry, and rollever if needed'''
        ret = LRDataSet.Add(self, spec, (peer, port), callback)
        if len(self.data) > self.max:
            self.data = self.data[-self.min:]
        return ret

